#import <simd/simd.h>
#import <ModelIO/ModelIO.h>
#include <vector>
#include <map>

#import "MetalViewDelegate.h"

// Include header shared between C code here, which executes Metal API commands, and .metal files
#import "ShaderTypes.h"

#include "../cpp/Renderer.h"
#include "IOSAssetManager.h"

static const NSUInteger MaxBuffersInFlight = 3;

@implementation MetalViewDelegate
{
    dispatch_semaphore_t _inFlightSemaphore;
    id <MTLDevice> _device;
    id <MTLCommandQueue> _commandQueue;
    
    id <MTLBuffer> _dynamicUniformBuffer[MaxBuffersInFlight];
    id <MTLRenderPipelineState> _pipelineState;
    id <MTLDepthStencilState> _depthState;
    id <MTLTexture> _dummyTexture;
    MTLVertexDescriptor *_mtlVertexDescriptor;
    
    vector_float2 _resolution;
    
    uint8_t _bufferIndex;
    
    id<MTLBuffer> _vertexBuffers[MaxBuffersInFlight];
    // In number of rects
    size_t _vertexBufferCapacities[MaxBuffersInFlight];
    
    std::vector<KDERect> _drawRects;
    std::vector<id<MTLTexture>> _drawTextures;
    
    std::map<KDEStringHandle, Texture> _textures;
}

-(nonnull instancetype)initWithMetalKitView:(nonnull MTKView *)view;
{
    self = [super init];
    if(self)
    {
        _device = view.device;
        _inFlightSemaphore = dispatch_semaphore_create(MaxBuffersInFlight);
        [self _loadMetalWithView:view];
    }

    return self;
}

- (void)_loadMetalWithView:(nonnull MTKView *)view;
{
    /// Load Metal state objects and initialize renderer dependent view properties
    
    view.depthStencilPixelFormat = MTLPixelFormatDepth32Float_Stencil8;
    view.colorPixelFormat = MTLPixelFormatBGRA8Unorm_sRGB;
    view.sampleCount = 1;
    
    _resolution = vector2((float)[view drawableSize].width, (float)[view drawableSize].height);
    
    _mtlVertexDescriptor = [[MTLVertexDescriptor alloc] init];
    
    _mtlVertexDescriptor.attributes[VertexAttributeDstP0].format = MTLVertexFormatFloat2;
    _mtlVertexDescriptor.attributes[VertexAttributeDstP0].offset = 0;
    _mtlVertexDescriptor.attributes[VertexAttributeDstP0].bufferIndex = BufferIndexRects;
    
    _mtlVertexDescriptor.attributes[VertexAttributeDstP1].format = MTLVertexFormatFloat2;
    _mtlVertexDescriptor.attributes[VertexAttributeDstP1].offset = offsetof(KDERect, dst_p1_x);
    _mtlVertexDescriptor.attributes[VertexAttributeDstP1].bufferIndex = BufferIndexRects;
    
    _mtlVertexDescriptor.attributes[VertexAttributeSrcP0].format = MTLVertexFormatFloat2;
    _mtlVertexDescriptor.attributes[VertexAttributeSrcP0].offset = offsetof(KDERect, src_p0_x);
    _mtlVertexDescriptor.attributes[VertexAttributeSrcP0].bufferIndex = BufferIndexRects;
    
    _mtlVertexDescriptor.attributes[VertexAttributeSrcP1].format = MTLVertexFormatFloat2;
    _mtlVertexDescriptor.attributes[VertexAttributeSrcP1].offset = offsetof(KDERect, src_p1_x);
    _mtlVertexDescriptor.attributes[VertexAttributeSrcP1].bufferIndex = BufferIndexRects;
    
    _mtlVertexDescriptor.attributes[VertexAttributeColor].format = MTLVertexFormatFloat4;
    _mtlVertexDescriptor.attributes[VertexAttributeColor].offset = offsetof(KDERect, r);
    _mtlVertexDescriptor.attributes[VertexAttributeColor].bufferIndex = BufferIndexRects;
    
    _mtlVertexDescriptor.attributes[VertexAttributeExpandR].format = MTLVertexFormatInt;
    _mtlVertexDescriptor.attributes[VertexAttributeExpandR].offset = offsetof(KDERect, expand_r);
    _mtlVertexDescriptor.attributes[VertexAttributeExpandR].bufferIndex = BufferIndexRects;
    
    _mtlVertexDescriptor.layouts[BufferIndexRects].stride = sizeof(KDERect);
    _mtlVertexDescriptor.layouts[BufferIndexRects].stepRate = 1;
    _mtlVertexDescriptor.layouts[BufferIndexRects].stepFunction = MTLVertexStepFunctionPerInstance;
    
    id<MTLLibrary> defaultLibrary = [_device newDefaultLibrary];
    id <MTLFunction> vertexFunction = [defaultLibrary newFunctionWithName:@"vertexShader"];
    id <MTLFunction> fragmentFunction = [defaultLibrary newFunctionWithName:@"fragmentShader"];
    
    MTLRenderPipelineDescriptor *pipelineStateDescriptor = [[MTLRenderPipelineDescriptor alloc] init];
    pipelineStateDescriptor.label = @"MyPipeline";
    pipelineStateDescriptor.rasterSampleCount = view.sampleCount;
    pipelineStateDescriptor.vertexFunction = vertexFunction;
    pipelineStateDescriptor.fragmentFunction = fragmentFunction;
    pipelineStateDescriptor.vertexDescriptor = _mtlVertexDescriptor;
    pipelineStateDescriptor.colorAttachments[0].pixelFormat = view.colorPixelFormat;
    pipelineStateDescriptor.depthAttachmentPixelFormat = view.depthStencilPixelFormat;
    pipelineStateDescriptor.stencilAttachmentPixelFormat = view.depthStencilPixelFormat;
    
    NSError *error = NULL;
    _pipelineState = [_device newRenderPipelineStateWithDescriptor:pipelineStateDescriptor error:&error];
    if (!_pipelineState)
    {
        NSLog(@"Failed to created pipeline state, error %@", error);
    }
    
    MTLDepthStencilDescriptor *depthStateDesc = [[MTLDepthStencilDescriptor alloc] init];
    depthStateDesc.depthCompareFunction = MTLCompareFunctionLess;
    depthStateDesc.depthWriteEnabled = NO;
    _depthState = [_device newDepthStencilStateWithDescriptor:depthStateDesc];
    
    for(NSUInteger i = 0; i < MaxBuffersInFlight; i++)
    {
        _dynamicUniformBuffer[i] = [_device newBufferWithLength:sizeof(Uniforms)
                                                        options:MTLResourceStorageModeShared];
        
        _dynamicUniformBuffer[i].label = @"UniformBuffer";
        
        _vertexBufferCapacities[i] = 256;
        _vertexBuffers[i] = [_device newBufferWithLength:sizeof(Rect) * _vertexBufferCapacities[i]
                                                 options:MTLResourceStorageModeShared];
    }
    
    _commandQueue = [_device newCommandQueue];
    
    // Create the dummy texture (1x1 px white)
    uint8_t bytes[4] = { 255, 255, 255, 255 };
    _dummyTexture = [self createTexture:bytes width:1 height:1 format:MTLPixelFormatRGBA8Uint];
}

- (NSUInteger)_updateGameState
{
    // Demo code
    static float x = 1.f;
    static float d = -0.01f;
    x += d;
    if (x <= 0.f)
        d *= -1.f;
    else if (x >= 1.f)
        d *= -1.f;

#if 0
    if (input_event_count > 0) {
        m_smiley_pos = input_events[input_event_count - 1].end;
        m_smiley_pos.x -= 250;
        m_smiley_pos.y -= 250;
    }
#endif

    Renderer::ptr->addRect(100, 100, 500, 500, 1.f, 1.f, 1.f, 1.f, StringRepository::global->internString("smiley_PNG42.png"));
    //Renderer::ptr->addRect(m_smiley_pos.x, m_smiley_pos.y, 500, 500, 0.f, x * x, 1.f - x * x, 1.f, m_smiley);
    
    // Upload rects
    if (_drawRects.size() > _vertexBufferCapacities[_bufferIndex]) {
        size_t newCap = ((_drawRects.size() + 63) / 64) * 64;
        _vertexBuffers[_bufferIndex] = [_device newBufferWithLength:sizeof(KDERect) * newCap
                                                                   options:MTLResourceStorageModeShared];
        _vertexBufferCapacities[_bufferIndex] = newCap;
    }
    memcpy(_vertexBuffers[_bufferIndex].contents,
           _drawRects.data(),
           _drawRects.size() * sizeof(KDERect));
    NSUInteger rectCount = (NSUInteger)_drawRects.size();
    _drawRects.clear();
    
    /// Update any game state before encoding renderint commands to our drawable
    Uniforms * uniforms = (Uniforms*)_dynamicUniformBuffer[_bufferIndex].contents;
    uniforms->resolution = _resolution;
    
    return rectCount;
}

- (void)drawInMTKView:(nonnull MTKView *)view
{
    /// Per frame updates here

    dispatch_semaphore_wait(_inFlightSemaphore, DISPATCH_TIME_FOREVER);

    _bufferIndex = (_bufferIndex + 1) % MaxBuffersInFlight;

    id <MTLCommandBuffer> commandBuffer = [_commandQueue commandBuffer];
    commandBuffer.label = @"MyCommand";

    __block dispatch_semaphore_t block_sema = _inFlightSemaphore;
    [commandBuffer addCompletedHandler:^(id<MTLCommandBuffer> buffer)
     {
         dispatch_semaphore_signal(block_sema);
     }];

    NSUInteger rectCount = [self _updateGameState];
    
    /// Delay getting the currentRenderPassDescriptor until absolutely needed. This avoids
    ///   holding onto the drawable and blocking the display pipeline any longer than necessary
    MTLRenderPassDescriptor* renderPassDescriptor = view.currentRenderPassDescriptor;

    if(renderPassDescriptor != nil)
    {
        /// Final pass rendering code here
        id <MTLRenderCommandEncoder> renderEncoder =
        [commandBuffer renderCommandEncoderWithDescriptor:renderPassDescriptor];
        renderEncoder.label = @"MyRenderEncoder";

        [renderEncoder pushDebugGroup:@"DrawRects"];

        [renderEncoder setFrontFacingWinding:MTLWindingCounterClockwise];
        [renderEncoder setCullMode:MTLCullModeNone];
        [renderEncoder setRenderPipelineState:_pipelineState];
        [renderEncoder setDepthStencilState:_depthState];

        [renderEncoder setVertexBuffer:_dynamicUniformBuffer[_bufferIndex]
                                offset:0
                               atIndex:BufferIndexUniforms];

        [renderEncoder setFragmentBuffer:_dynamicUniformBuffer[_bufferIndex]
                                  offset:0
                                 atIndex:BufferIndexUniforms];

        [renderEncoder setVertexBuffer:_vertexBuffers[_bufferIndex]
                                offset:0
                               atIndex:BufferIndexRects];

        size_t start = 0;
        id<MTLTexture> currentTexture = _drawTextures[0];
        while (start < rectCount) {
            for (size_t i = start; i < rectCount; ++i) {
                id<MTLTexture> texture = _drawTextures[i];
                if (texture != currentTexture) {
                    size_t count = i - start;
                    [renderEncoder setFragmentTexture:currentTexture
                                              atIndex:TextureIndexColor];
                    
                    [renderEncoder drawPrimitives:MTLPrimitiveTypeTriangle
                                      vertexStart:0
                                      vertexCount:6
                                    instanceCount:count
                                     baseInstance:start];
                    start = i;
                    currentTexture = texture;
                    break;
                }
                else if (i == rectCount - 1) {
                    // just draw it
                    const size_t count = rectCount - start;
                    [renderEncoder setFragmentTexture:currentTexture
                                              atIndex:TextureIndexColor];
                    
                    [renderEncoder drawPrimitives:MTLPrimitiveTypeTriangle
                                      vertexStart:0
                                      vertexCount:6
                                    instanceCount:count
                                     baseInstance:start];
                    start = rectCount;
                    break;
                }
            }
        }
            
        [renderEncoder popDebugGroup];

        [renderEncoder endEncoding];

        [commandBuffer presentDrawable:view.currentDrawable];
    }
    _drawTextures.clear();

    [commandBuffer commit];
}

- (void)mtkView:(nonnull MTKView *)view drawableSizeWillChange:(CGSize)size
{
    /// Respond to drawable size or orientation changes here
    _resolution.x = (float)size.width;
    _resolution.y = (float)size.height;
}

- (void)pushRect:(KDERect)rect withTexture:(KDEStringHandle)texture
{
    Texture tex;
    if (texture != 0) {
        if (_textures.find(texture) == _textures.end()) {
            if (!AssetManager::ptr->loadTexture(StringRepository::global->getString(texture), &tex)) {
                NSLog(@"Failed to load a texture");
                return;
            }
            _textures.insert(std::make_pair(texture, tex));
        }
        else {
            tex = _textures[texture];
        }
        _drawTextures.push_back((__bridge id)tex.getObj());
    }
    else {
        _drawTextures.push_back(_dummyTexture);
    }
    _drawRects.push_back(rect);
}


- (void)pushRect:(KDERect)rect withFont:(KDEStringHandle)font withCharHeight:(float)char_height withKey:(KDEStringHandle)key
{
    Texture tex;
    if (key != 0) {
        if (_textures.find(key) == _textures.end()) {
            if (!AssetManager::ptr->loadFontBitmap(StringRepository::global->getString(font), char_height, &tex)) {
                NSLog(@"Failed to load font");
                return;
            }
            _textures.insert(std::make_pair(key, tex));
        }
        else {
            tex = _textures[key];
        }
        _drawTextures.push_back((__bridge id)tex.getObj());
    }
    else {
        _drawTextures.push_back(_dummyTexture);
    }
    _drawRects.push_back(rect);
}


-(id<MTLTexture>)createTexture:(nonnull const void*)data width:(unsigned int)width height:(unsigned int)height format:(MTLPixelFormat)format;
{
    MTLTextureDescriptor* desc = [MTLTextureDescriptor texture2DDescriptorWithPixelFormat:format width:(NSUInteger)width height:(NSUInteger)height mipmapped:NO];
    
    MTLRegion region = { {0, 0, 0}, {width, height, 1}};
    
    id<MTLTexture> texture = [_device newTextureWithDescriptor: desc];
    NSUInteger bytesPerRow = 4 * width;
    [texture replaceRegion:region mipmapLevel:0 withBytes:data bytesPerRow:bytesPerRow];
    return texture;
}

@end
